/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
/******************************************************************************
 *  utils.c   common utilities                                                *
 ******************************************************************************/
#include "vector.h"
#include "units.h"

short ShortBits(short s)
{
    short count = 0;
    for (; s; s = s << 1)
	if (s & 0x8000)
	    count++;
    return count;
}

long SubMod(long lon, long ctr)
{
    lon -= ctr;
    if (lon < -DEG180_LU)
	lon += DEG360_LU;
    if (lon > DEG180_LU)
	lon -= DEG360_LU;
    return lon;
}

int CenterMBR(MBR * rect)
{
    rect->WLon -= DEG180_LU;
    rect->ELon -= DEG180_LU;
    rect->SLat -= DEG90_LU;
    rect->NLat -= DEG90_LU;
}

char OverlapMBR(MBR rect1, MBR rect2)
{
    //fprintf(stderr, "r1.nlat=%ld r1.wlon=%ld r1.slat=%ld r1.elon=%ld\n",
    //      rect1.NLat, rect1.WLon, rect1.SLat, rect1.ELon);
    //fprintf(stderr, "r2.nlat=%ld r2.wlon=%ld r2.slat=%ld r2.elon=%ld\n",
    //      rect2.NLat, rect2.WLon, rect2.SLat, rect2.ELon);

    if ((rect1.SLat >= rect2.NLat) || (rect2.SLat >= rect1.NLat))
	return FALSE;
    if ((DATELN(rect1) + DATELN(rect2) > 0) && ((rect2.WLon > rect1.ELon) ||
					      (rect1.WLon > rect2.ELon)))
	return FALSE;
    /* neither cross 180, reject if W>e (type IV.) or w>E (III.) */
    if ((DATELN(rect1) * DATELN(rect2) < 0) && ((rect2.WLon > rect1.ELon) &&
					      (rect1.WLon > rect2.ELon)))
	return FALSE;
    /* only one crosses 180, reject if W>e and w>E (type Ia & IIb) */
    return TRUE;
    /* all other cases overlap */
}

/* is rect1 completely inside rect2? */
char RectInRect(MBR rect1, MBR rect2)
{
    if ((rect2.SLat > rect1.SLat) || (rect1.NLat > rect2.NLat))
	return FALSE;
    if ((DATELN(rect1) * DATELN(rect2) > 0) && ((rect1.WLon > rect2.WLon) &&
					      (rect2.ELon > rect1.ELon)))
	return TRUE;
    /* neither or both cross 180, pass if w>W and E>e (type I. & IIab) */
    else if ((DATELN(rect2) < 0) && ((rect1.WLon > rect2.WLon) ||
				     (rect2.ELon > rect1.ELon)))
	return TRUE;
    /* rect2 crosses, pass if w>W (type IIIa) or E>e (type  IVa) */
    else
	return FALSE;
    /* reject all other cases */
}

char PtInRect(long x, long y, MBR rect)
{
    return ((((DATELN(rect) > 0) && (x > rect.WLon) && (x < rect.ELon)) ||
	 ((DATELN(rect) < 0) && ((x > rect.WLon) || (x < rect.ELon)))) &&
	    (y >= rect.SLat) && (y <= rect.NLat));
}

/* determines "sector" of point WRT a rectangle, represented by a bitmask:
   0(1)          1(2)
   :      :
   2(4)   5   :  4   :  6
   ......:______:......
   |      |
   1   |  0   |  2
   ......|______|......
   3(8)   9   :  8   :  10
   :      :         */
char Sector(long x, long y, MBR rect)
{
    return ((x < rect.WLon) + ((x > rect.ELon) << 1) +
	    ((y > rect.NLat) << 2) + ((y < rect.SLat) << 3));
}

/* Search through a string of CFCC codes (w/ wildcards) for a particular code */
char CodeInList(char *code, char *list)
{
    int i;
    for (i = 0; i < strlen(list); i += 3) {
	if ((*code == *(list + i)) || (*(list + i) == 42))
	    if ((*(code + 1) == *(list + i + 1)) || (*(list + i + 1) == 42))
		if ((*(code + 2) == *(list + i + 2)) || (*(list + i + 2) == 42))
		    return TRUE;
    }
    return FALSE;
}

/* Given a CFCC code, search through the feature definitions for a match, return symbology */
int FeatureSym(MapStruct * ms, Symbol ** linesym, Symbol ** fillsym, Symbol ** ptsym, char *code)
{
    int i, j;
    for (i = 0; i < ms->nfeats; i++) {
	if (CodeInList(code, ms->features[i].name)) {
	    if (ms->features[i].line.color != -1)
		*linesym = &ms->features[i].line;
	    else
		*linesym = NULL;
	    if (ms->features[i].fill.color != -1)
		*fillsym = &ms->features[i].fill;
	    else
		*fillsym = NULL;
	    if (ms->features[i].point.color != -1)
		*ptsym = &ms->features[i].point;
	    else
		*ptsym = NULL;
	    return i + 1;
	}
    }
    return FALSE;
}

